/*-
 * Copyright (c) 1988, 1989 The Regents of the University of California.
 * All rights reserved.
 * Copyright (c) 2008, 2017 Pengutronix e.K.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the University of
 *        California, Berkeley, Pengutronix e.K., and their contributors.
 * 4. Neither the name of the copyright holders nor the names of their
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * remote shell server:
 *	[port]\0
 *	remuser\0
 *	locuser\0
 *	command\0
 *	data
 */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <pwd.h>
#include <grp.h>
#include <syslog.h>
#include <resolv.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>  /* for vsnprintf */
#include <stdlib.h>
#include <string.h>
#include <paths.h>
#include <stdarg.h>
#include <ctype.h>
#include <assert.h>
#include <sched.h>
#include <getopt.h>

#include <urshd_config.h>

#define PROG "urshd"

#define	OPTIONS	"hnp:P:D"

#define USAGE									\
"Usage: " PROG " [options]\n"							\
"Options:\n\n"									\
"  -h\n"									\
"  --help			Display this message\n\n"			\
"  -n\n"									\
"  --disable-keepalive		Do not set SO_KEEPALIVE socket option\n\n"	\
"  -p <num>\n"									\
"  --priority=<num>		Set " PROG " priority to <num>\n\n"		\
"  -P <num>\n"									\
"  --command-priority=<num>	Set command priority to <num>\n\n"		\
"  -D				don't daemonize\n\n"

#define STDERR 1
#define SYSLOG 2

static struct option long_options[] = {
	{ "help", no_argument, NULL, 'h' },
	{ "disable-keepalive", no_argument, NULL, 'n' },
	{ "priority", required_argument, NULL, 'p' },
	{ "command-priority", required_argument, NULL, 'P' },
	{ NULL, 0, NULL, 0}
};

static int keepalive = 1;
static int sent_null;
static int priority;
static int cmd_priority;
static int opt_daemon = 1;

static char username[20] = "USER=";
static char homedir[64] = "HOME=";
static char shell[64] = "SHELL=";
static char path[100] = "PATH=";
static char *envinit[] = {homedir, shell, path, username, NULL};
extern char **environ;

static void error(const char *fmt, ...);
static void doit(struct sockaddr_in6 *fromp);
static void getstr(char *buf, int cnt, const char *err);

static void
set_priority(int sched_priority)
{
	struct sched_param schedp = {
		.sched_priority = sched_priority,
	};

	sched_setscheduler(0, SCHED_FIFO, &schedp);
}

static void
usage(int where, char *error, ...)
{
	va_list ap;

	if ((where & STDERR) != 0) {
		if (error) {
			fprintf(stderr, PROG ": ");

			va_start(ap, error);
			vfprintf(stderr, error, ap);
			va_end(ap);

			fprintf(stderr, "\n\n");
		}

		fprintf(stderr, USAGE);
	}

	if ((where & SYSLOG) != 0) {
		if (error) {
			va_start(ap, error);
			vsyslog(LOG_ERR, error, ap);
			va_end(ap);
		}

		/*syslog(LOG_ERR, USAGE);*/
	}

	exit(2);
}

static void
fatal(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vsyslog(LOG_ERR, fmt, ap);
	va_end(ap);

	exit(1);
}

/*
 * Report error to client.
 * Note: can't be used until second socket has connected
 * to client, or older clients will hang waiting
 * for that connection first.
 */
static void
error(const char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZ], *bp = buf;

	if (sent_null == 0) {
		*bp++ = 1;
	}

	va_start(ap, fmt);
	vsnprintf(bp, sizeof(buf) - 1, fmt, ap);
	va_end(ap);

	if (write(2, buf, strlen(buf)) < 0) {
		fatal("error: write: %m");
	}
}

static void
getstr(char *buf, int cnt, const char *err)
{
	char c;
	do {
		if (read(0, &c, 1) != 1) exit(1);
		*buf++ = c;
		if (--cnt == 0) {
			error("%s too long\n", err);
			exit(1);
		}
	} while (c != 0);
}

static int
getint(void)
{
	int port = 0;
	char c;
	do {
		if (read(0, &c, 1) != 1)
			exit(1);
		if (isascii(c) && isdigit(c))
			port = port*10 + c-'0';
	} while (c != 0);
	return port;
}

static void
stderr_parent(int sock, int pype, int pid)
{
	fd_set ready, readfrom;
	char buf[BUFSIZ], sig;
	int one = 1;
	int nfd, cc, guys=2;

	ioctl(pype, FIONBIO, (char *)&one);
	/* should set s nbio! */

	FD_ZERO(&readfrom);
	FD_SET(sock, &readfrom);
	FD_SET(pype, &readfrom);

	if (pype > sock)
		nfd = pype+1;
	else
		nfd = sock+1;

	while (guys > 0) {
		ready = readfrom;
		if (select(nfd, &ready, NULL, NULL, NULL) < 0) {
			if (errno != EINTR) {
				break;
			}
			continue;
		}
		if (FD_ISSET(sock, &ready)) {
			cc = read(sock, &sig, 1);
			if (cc <= 0) {
				FD_CLR(sock, &readfrom);
				guys--;
			} else
				killpg(pid, sig);
		}
		if (FD_ISSET(pype, &ready)) {
			cc = read(pype, buf, sizeof(buf));
			if (cc <= 0) {
				shutdown(sock, 2);
				FD_CLR(pype, &readfrom);
				guys--;
			} else {
				if (write(sock, buf, cc) < 0) {
					fatal("stderr_parent: %m");
				}
			}
		}
	}

	close(sock);
	close(pype);

	exit(0);
}

static void
doit(struct sockaddr_in6 *fromp)
{
	char cmdbuf[sysconf(_SC_ARG_MAX) + 1];
	const char *theshell, *shellname;
	char locuser[16], remuser[16];
	struct passwd *pwd;
	int sock = -1;
	u_short port;
	int pv[2], pid, ifd;

	signal(SIGINT, SIG_DFL);
	signal(SIGQUIT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);

	alarm(60);
	port = getint();
	alarm(0);

	if (port != 0) {
		int lport = IPPORT_RESERVED - 1;
		sock = rresvport_af(&lport, AF_INET6);
		if (sock < 0) {
			syslog(LOG_ERR, "can't get stderr port: %m");
			exit(1);
		}
		if (port >= IPPORT_RESERVED) {
			syslog(LOG_ERR, "2nd port not reserved\n");
			exit(1);
		}
		fromp->sin6_port = htons(port);
		if (connect(sock, (struct sockaddr *)fromp,
		            sizeof(*fromp)) < 0) {
			syslog(LOG_INFO, "connect second port: %m");
			exit(1);
		}
	}

	getstr(remuser, sizeof(remuser), "remuser");
	getstr(locuser, sizeof(locuser), "locuser");
	getstr(cmdbuf, sizeof(cmdbuf), "command");

	setpwent();
	pwd = getpwnam(locuser);
	if (pwd == NULL) {
		syslog(LOG_INFO|LOG_AUTH, PROG " permission denied to %s as %s.",
		       remuser, locuser);
		syslog(LOG_INFO|LOG_AUTH, PROG " command was '%s'", cmdbuf);
		error("Permission denied.\n");
		exit(1);
	}

	if (chdir(pwd->pw_dir) < 0) {
		if (chdir("/") != 0) {
			error("No remote directory.\n");
			exit(1);
		}
	}

	if (pwd->pw_uid != 0 && !access(_PATH_NOLOGIN, F_OK)) {
		error("Logins currently disabled.\n");
		exit(1);
	}

	if (write(2, "\0", 1) < 0) {
		fatal("Unable to write: %m");
	}
	sent_null = 1;

	if (port) {
		if (pipe(pv) < 0) {
			error("Can't make pipe.\n");
			exit(1);
		}
		pid = fork();
		if (pid == -1)  {
			error("Can't fork; try again.\n");
			exit(1);
		}
		if (pid) {
			close(0);
			close(1);
			close(2);
			close(pv[1]);
			stderr_parent(sock, pv[0], pid);
			/* NOTREACHED */
		}
		setpgrp();
		close(sock);
		close(pv[0]);
		dup2(pv[1], 2);
		close(pv[1]);
	}

	theshell = pwd->pw_shell;
	if (!theshell || !*theshell) {
		/* shouldn't we deny access? */
		theshell = _PATH_BSHELL;
	}

#if BSD > 43
	if (setlogin(pwd->pw_name) < 0) {
		syslog(LOG_ERR, "setlogin() failed: %m");
	}
#endif
	if (setgid(pwd->pw_gid)) {
		syslog(LOG_ERR, "setgid: %m");
		exit(1);
	}
	if (initgroups(pwd->pw_name, pwd->pw_gid)) {
		syslog(LOG_ERR, "initgroups: %m");
		exit(1);
	}
	if (setuid(pwd->pw_uid)) {
		syslog(LOG_ERR, "setuid: %m");
		exit(1);
	}
	environ = envinit;

	strncat(homedir, pwd->pw_dir, sizeof(homedir)-6);
	homedir[sizeof(homedir)-1] = 0;

	strcat(path, _PATH_DEFPATH);

	strncat(shell, theshell, sizeof(shell)-7);
	shell[sizeof(shell)-1] = 0;

	strncat(username, pwd->pw_name, sizeof(username)-6);
	username[sizeof(username)-1] = 0;

	shellname = strrchr(theshell, '/');
	if (shellname)
		shellname++;
	else
		shellname = theshell;

	endpwent();

	if (0) {
		syslog(LOG_INFO|LOG_AUTH, "%s as %s: cmd='%s'", remuser, locuser, cmdbuf);
	}

	/*
	 * Close all fds, in case libc has left fun stuff like
	 * /etc/shadow open.
	 */
	for (ifd = getdtablesize()-1; ifd > 2; ifd--)
		close(ifd);

	if (cmd_priority) {
		set_priority(cmd_priority);
	}

	execl(theshell, shellname, "-c", cmdbuf, NULL);
	perror(theshell);
	exit(1);
}

static void
network_init(int fd, struct sockaddr_in6 *fromp)
{
	struct linger linger;
	socklen_t fromlen;
	int on=1;

	fromlen = sizeof(*fromp);
	if (getpeername(fd, (struct sockaddr *) fromp, &fromlen) < 0) {
		syslog(LOG_ERR, "getpeername: %m");
		_exit(1);
	}
	if (keepalive &&
	    setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *)&on,
	    sizeof(on)) < 0)
		syslog(LOG_WARNING, "setsockopt (SO_KEEPALIVE): %m");
	linger.l_onoff = 1;
	linger.l_linger = 60;			/* XXX */
	if (setsockopt(fd, SOL_SOCKET, SO_LINGER, (char *)&linger,
	    sizeof (linger)) < 0)
		syslog(LOG_WARNING, "setsockopt (SO_LINGER): %m");

#ifdef IP_OPTIONS
      {
	u_char optbuf[BUFSIZ/3], *cp;
	char lbuf[BUFSIZ+1], *lp;
	char peeraddr[5*8+8];   // 8 groups of "xxxx:" plus some slack
	socklen_t optsize = sizeof(optbuf);
	int  ipproto;
	struct protoent *ip;

	if ((ip = getprotobyname("ipv6")) != NULL)
		ipproto = ip->p_proto;
	else
		ipproto = IPPROTO_IPV6;
	if (!getsockopt(0, ipproto, IP_OPTIONS, (char *)optbuf, &optsize) &&
	    optsize != 0) {
		lp = lbuf;

		/*
		 * If these are true, this will not run off the end of lbuf[].
		 */
		assert(optsize <= BUFSIZ/3);
		assert(3*optsize <= BUFSIZ);
		for (cp = optbuf; optsize > 0; cp++, optsize--, lp += 3)
			snprintf(lp, 4, " %2.2x", *cp);

		inet_ntop(AF_INET6, &(fromp->sin6_addr),
				peeraddr, sizeof(peeraddr));
		syslog(LOG_NOTICE,
		       "Connection received from %s using IP options"
		       " (ignored): %s",
		       peeraddr, lbuf);

		/* These lines go back to the oldest netkit-rsh I could find.
		 * According to the internet, this should clear source routing
		 * options in the incoming packet. I don't know why this is
		 * necessary, but for IPv6, there is no corresponding option.
		 *
		if (setsockopt(0, ipproto, IP_OPTIONS, NULL, optsize) != 0) {
			syslog(LOG_ERR, "setsockopt IP_OPTIONS NULL: %m");
			exit(1);
		}
		*/
	}
      }
#endif
}

static void
mainloop()
{
	int s;
	struct sockaddr_in6 sn;
	int on = 1;
	int off = 0;
	struct addrinfo * ai, hints = {
		.ai_family = AF_INET6,
		.ai_socktype = SOCK_STREAM,
		.ai_flags = AI_PASSIVE,
	};

	signal(SIGCHLD, SIG_IGN);

	if (priority) {
		set_priority(priority);
	}

	if (opt_daemon) {
		if (daemon(0, 0) != 0) {
			fatal("Could not fork: %s", strerror(errno));
		}
	}

	s = getaddrinfo(NULL, "shell", &hints, &ai);
	if (s != 0) {
		freeaddrinfo(ai);
		fatal("getaddrinfo: %s", gai_strerror(s));
	}

	s = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
	if (s < 0) {
		freeaddrinfo(ai);
		fatal("socket: %s", strerror(errno));
	}
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) != 0) {
		freeaddrinfo(ai);
		syslog(LOG_WARNING, "setsockopt (SO_REUSEADDR): %m");
	}
	if (setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY, &off, sizeof(off)) != 0) {
		freeaddrinfo(ai);
		fatal("setsockopt (IPV6_V6ONLY): %s", strerror(errno));
	}
	if (bind(s, ai->ai_addr, ai->ai_addrlen) < 0) {
		freeaddrinfo(ai);
		fatal("bind: %s", strerror(errno));
	}
	if (listen(s, 1) < 0) {
		freeaddrinfo(ai);
		fatal("listen: %s", strerror(errno));
	}
	freeaddrinfo(ai);
	while (1) {
		socklen_t len;
		int ns, pid;
		struct sockaddr_in6 from;

		syslog(LOG_INFO, "Waiting for connections.");

		len = sizeof(sn);
		ns = accept(s, (struct sockaddr *)&sn, &len);
		if (ns < 0) {
			fatal("accept: %s", strerror(errno));
		}

		if ((pid = fork()) < 0) {
			fatal("fork: %s", strerror(errno));
		} else if (pid == 0) {
			close(s);
			dup2(ns, 0);
			dup2(ns, 1);
			dup2(ns, 2);
			close(ns);
			network_init(0, &from);
			doit(&from);
		} else {
			close(ns);
		}
	}
}

int
main(int argc, char *argv[])
{
	int ch;
	int index;

	if (0 != geteuid()) {
		usage(STDERR, "super-user effective user ID is expected");
	}

	openlog(PROG, LOG_PID | LOG_ODELAY, LOG_DAEMON);

	opterr = 0;
	while ((ch = getopt_long(argc, argv, OPTIONS, long_options, &index)) != EOF) {
		switch (ch) {
		case 'n':
			keepalive = 0;
			break;

		case 'p':
			priority = atoi(optarg);
			break;

		case 'P':
			cmd_priority = atoi(optarg);
			break;

		case 'h':
			usage(STDERR, NULL);
			break;

		case 'D':
			opt_daemon = 0;
			break;

		default:
			if (optopt == 0) {
				usage(STDERR | SYSLOG, "unrecognized command line option: %s", argv[optind - 1]);
			} else {
				usage(STDERR | SYSLOG, "argument to -%c option is missing", optopt);
			}
		}
	}

	if (optind < argc) {
		usage(STDERR | SYSLOG, "Unexpected argument: %s", argv[optind]);
	}

	mainloop();
	return 0;
}
/* vim: set ft=c ts=8 noet */
